import { Effect, Option, Sink, Stream } from "effect";
import { RawAccount } from "cojson";
import { Account, AccountCtx, Group, SubscriptionScope, Ref, inspect, subscriptionsScopes, } from "../internal.js";
import { fulfillsDepth } from "./deepLoading.js";
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function isCoValue(value) {
    return value && value._type !== undefined;
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function isCoValueClass(value) {
    return typeof value === "function" && value.fromRaw !== undefined;
}
/** @internal */
export class CoValueBase {
    get _owner() {
        const owner = this._raw.group instanceof RawAccount
            ? Account.fromRaw(this._raw.group)
            : Group.fromRaw(this._raw.group);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subScope.onRefAccessedOrSet(this.id, owner.id);
            subscriptionsScopes.set(owner, subScope);
        }
        return owner;
    }
    /** @private */
    get _loadedAs() {
        return Account.fromNode(this._raw.core.node);
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    constructor(..._args) {
        Object.defineProperty(this, "_instanceID", {
            value: `instance-${Math.random().toString(36).slice(2)}`,
            enumerable: false,
        });
    }
    /** @category Internals */
    static fromRaw(raw) {
        return new this({ fromRaw: raw });
    }
    static load(...args) {
        const { id, as, depth } = args.length === 3
            ? { id: args[0], as: args[1], depth: args[2] }
            : { id: args[0].id, as: args[0]._loadedAs, depth: args[1] };
        return Effect.runPromise(this.loadEf(id, depth).pipe(Effect.mapError(() => undefined), Effect.merge, Effect.provideService(AccountCtx, as)));
    }
    /** @category Subscription & Loading */
    static loadEf(id, depth) {
        return this.subscribeEf(id, depth).pipe(Stream.runHead, Effect.andThen(Effect.mapError((_noSuchElem) => "unavailable")));
    }
    static subscribe(...args) {
        const { id, as, depth, listener } = args.length === 4
            ? {
                id: args[0],
                as: args[1],
                depth: args[2],
                listener: args[3],
            }
            : {
                id: args[0].id,
                as: args[0]._loadedAs,
                depth: args[1],
                listener: args[2],
            };
        void Effect.runPromise(Effect.provideService(this.subscribeEf(id, depth).pipe(Stream.run(Sink.forEach((update) => Effect.sync(() => listener(update))))), AccountCtx, as));
        return function unsubscribe() { };
    }
    /** @category Subscription & Loading */
    static subscribeEf(id, depth) {
        return AccountCtx.pipe(Effect.andThen((account) => new Ref(id, account, this).loadEf()), Stream.fromEffect, Stream.flatMap((value) => Stream.asyncScoped((emit) => Effect.gen(this, function* (_) {
            const subscription = new SubscriptionScope(value, this, (update) => void emit.single(update));
            yield* _(Effect.addFinalizer(() => Effect.sync(() => subscription.unsubscribeAll())));
        }))), Stream.filterMap((update) => Option.fromNullable(fulfillsDepth(depth, update)
            ? update
            : undefined)));
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    toJSON() {
        return {
            id: this.id,
            type: this._type,
            error: "unknown CoValue class",
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Type Helpers*/
    as(otherSchema) {
        const cast = otherSchema.fromRaw(this._raw);
        const subScope = subscriptionsScopes.get(this);
        if (subScope) {
            subscriptionsScopes.set(cast, subScope);
        }
        return cast;
    }
}
//# sourceMappingURL=interfaces.js.map